/*
Copyright (c) 2004 Ivan Skytte Jrgensen

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the
use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.

3. This notice may not be removed or altered from any source distribution.
*/

#include <sys/pstat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/un.h>

static const char *pstate2string(int pstate) {
    switch(pstate) {
        case PS_TCPS_CLOSED: return "closed";
        case PS_TCPS_IDLE: return "idle";
        case PS_TCPS_BOUND: return "bound";
        case PS_TCPS_LISTEN: return "listen";
        case PS_TCPS_SYN_SENT: return "SYN_SENT";
        case PS_TCPS_SYN_RCVD: return "SYN_RCVD";
        case PS_TCPS_ESTABLISHED: return "established";
        case PS_TCPS_CLOSE_WAIT: return "close-wait";
        case PS_TCPS_FIN_WAIT_1: return "FIN-WAIT-1";
        case PS_TCPS_CLOSING: return "closing";
        case PS_TCPS_LAST_ACK: return "last-ack";
        case PS_TCPS_FIN_WAIT_2: return "FIN-WAIT-2";
        case PS_TCPS_TIME_WAIT: return "time-wait";
    }
    return "?";
}

static void print_protocol(uint32_t p) {
    switch(p) {
        case PS_PROTO_ICMP:
            printf("ICMP");
            break;
        case PS_PROTO_IGMP:
            printf("IGMP");
            break;
        case PS_PROTO_IPIP:
            printf("IP-IP");
            break;
        case PS_PROTO_TCP:
            printf("TCP");
            break;
        case PS_PROTO_UDP:
            printf("UDP");
            break;
        case PS_PROTO_ESP:
            printf("ESP");
            break;
        case PS_PROTO_AH:
            printf("AH");
            break;
        case PS_PROTO_RAW:
            printf("RAW");
            break;
        default: /*there are more but this is enough*/
            printf("protocol=%u",p);
    }
}

static void print_fileinfo2(/*const*/ pst_fileinfo2 *fi2) {
    printf("%5lu ", fi2->psf_fd);
    switch(fi2->psf_ftype) {
        case PS_TYPE_UNKNOWN:
            printf("unknown  ");
            break;
        case PS_TYPE_VNODE: {
            char pathname[1024];
            printf("file     ");
			if(pstat_getpathname(pathname,sizeof(pathname),&fi2->psf_fid)>0)
			    printf("%s",pathname);
            printf(" @%ld", fi2->_PSF_OFFSET64);
            break;
        }
        case PS_TYPE_SOCKET: {
            struct pst_socket psts;
            printf("socket   ");
            if(pstat_getsocket(&psts,sizeof(psts),&fi2->psf_fid)>0) {
                switch(psts.pst_family) {
                    case AF_UNIX: {
                        struct sockaddr_un sa;
                        printf("unix ");
                        printf(" ");
                        memcpy(&sa,psts.pst_boundaddr,sizeof(sa));
                        printf("%s",sa.sun_path);
                        printf(" -> ");
                        memcpy(&sa,psts.pst_remaddr,sizeof(sa));
                        printf("%s",sa.sun_path);
                        printf(" node:%llx",*(uint64_t*)(&psts.pst_peer_hi_nodeid));
                        break;
                    }
                    case AF_INET: {
                        printf("inet ");
                        print_protocol(psts.pst_protocol);
                        switch(psts.pst_protocol) {
                            case PS_PROTO_TCP:
                            case PS_PROTO_UDP: {
                                struct sockaddr_in sa;
                                char buf[64];
                                printf(" ");
                                memcpy(&sa,psts.pst_boundaddr,sizeof(sa));
                                printf("%s:%u", sa.sin_addr.s_addr==0?"*":inet_ntop(AF_INET,&sa.sin_addr,buf,sizeof(buf)),ntohs(sa.sin_port));
                                printf(" -> ");
                                memcpy(&sa,psts.pst_remaddr,sizeof(sa));
                                printf("%s:%u", sa.sin_addr.s_addr==0?"*":inet_ntop(AF_INET,&sa.sin_addr,buf,sizeof(buf)),ntohs(sa.sin_port));
                                printf(" %s",pstate2string((int)psts.pst_pstate));
                                break;
                            }
                        }
                        break;
                    }
                    case AF_INET6: {
                        printf("inet6 ");
                        print_protocol(psts.pst_protocol);
                        switch(psts.pst_protocol) {
                            case PS_PROTO_TCP:
                            case PS_PROTO_UDP: {
                                struct sockaddr_in6 sa;
                                char buf[64];
                                printf(" ");
                                memcpy(&sa,psts.pst_boundaddr,sizeof(sa));
                                printf("%s/%u", IN6_IS_ADDR_UNSPECIFIED(&sa.sin6_addr)?"*":inet_ntop(AF_INET6,&sa.sin6_addr,buf,sizeof(buf)),ntohs(sa.sin6_port));
                                printf(" -> ");
                                memcpy(&sa,psts.pst_remaddr,sizeof(sa));
                                printf("%s/%u", IN6_IS_ADDR_UNSPECIFIED(&sa.sin6_addr)?"*":inet_ntop(AF_INET6,&sa.sin6_addr,buf,sizeof(buf)),ntohs(sa.sin6_port));
                                printf(" %s",pstate2string((int)psts.pst_pstate));
                                break;
                            }
                        }
                        break;
                    }
                    default:
                        printf("family=%u",psts.pst_family);
                }
            }
            break;
        }
        case PS_TYPE_STREAMS: {
            char pathname[1024];
            struct pst_stream psts[2];
            int rc;
            int i;
            printf("stream   ");
            
			if(pstat_getpathname(pathname,sizeof(pathname),&fi2->psf_fid)>0)
			    printf("%s ",pathname);
            
            rc=pstat_getstream(psts,sizeof(psts[0]),2,0,&fi2->psf_fid);
            for(i=0; i<rc; i++) {
                printf("[");
                switch(psts[i].type) {
                    case PS_STR_HEAD: {
                        printf("head ");
                        printf("major:%ld minor:%ld",psts[i].val.head.pst_dev_major,psts[i].val.head.pst_dev_minor);
                        break;
                    }
                    case PS_STR_MODULE: {
                        printf("module ");
                        printf("%s", psts[i].val.module.pst_name);
                        break;
                    }
                    case PS_STR_DRIVER: {
                        printf("driver ");
                        printf("%s", psts[i].val.driver.pst_name);
                        break;
                    }
                }
                printf("] ");
            }
            break;
        }
        case PS_TYPE_UNSP:
            printf("user-nsp ");
            break;
        case PS_TYPE_LLA:
            printf("l-l-lan  ");
            break;
        default:
            printf("?        ");
    }
    printf("\n");
}
/*
      int pstat_getstream(
           struct pst_stream *buf, size_t elemsize, size_t elemcount,
           int moduleskip, struct pst_fid *fid
      );
*/

int main(int argc, char **argv) {
    pid_t pid;
    int index;
    char *endptr=0;
    
    if(argc!=2 || strcmp(argv[1],"-?")==0 || strcmp(argv[1],"--usage")==0) {
        fprintf(stderr,"pfiles - display open files in a process\n");
        fprintf(stderr,"usage: pfiles pid\n");
        fprintf(stderr,"Dispalys all open file descriptors in a process, and\n");
        fprintf(stderr,"details for each descriptor.\n");
        return 99;
    }
    
    pid = (pid_t)strtoul(argv[1],&endptr,0);
    if(endptr && *endptr) {
        fprintf(stderr,"Could not parse '%s' as a process-id\n",argv[1]);
        return 99;
    }
    
    index=0;
    for(;;) {
#define NUM_FI2 32
        struct pst_fileinfo2 fi2[NUM_FI2];
        int rc = pstat_getfile2(fi2, sizeof(fi2[0]), NUM_FI2, index, pid);
        int i;
        if(rc<0) {
            perror("pfiles");
            return 1;
        }
        if(rc==0) break; /*finished*/
        
        for(i=0; i<rc; i++)
            print_fileinfo2(fi2+i);
        index = (int)(fi2[rc-1].psf_fd+1);
    }
    return 0;
}
